// File:      App.cpp
// Author:    Philip Rideout
// Brief:     Application class.

#include "App.h"
#include "Vector.h"
#include "Platform.h"
#include "MeshEggs.h"
#include <cmath>
#include <cassert>
#include <vector>
#include <hash_map>
#include <hash_set>

typedef std::vector<vec3> NormalList;
typedef std::vector<unsigned int> AdjacencyList;
typedef stdext::hash_set<unsigned int> AdjacencySet;

// TODO what distinguishes a vertex from a point? why the indirection? only because of LWO format?
//      is only because we want the same point to have different lightingNormals?
//      maybe that'll never happen because we compute lighting normals ourselves!
//      does LWO have lighting normals in the file itself?

struct Point
{
    vec3 position;
    AdjacencyList adjacentFaces;
    AdjacencySet adjacentVerts;
    vec3 faceCentroidsSum;
};

struct Vertex
{
    unsigned int index;
    vec3 lightingNormal;
};

struct Face
{
    Vertex vertices[4]; // TODO: allow non-quads
    vec3 facetNormal;
};

typedef std::vector<Point> PointList;
typedef std::vector<Face> FaceList;

struct Mesh
{
    PointList points;
    FaceList faces; // TODO: change to a LOD-index array
    unsigned int edgeCount;
};

class CApp : public IApp
{
  public:
    CApp();
    void Update();
    void Draw() const;
    void OnKey(char);
  protected:
    void initGL() const;
    void createMesh(Mesh*, const MeshEgg&) const;
    void subdivideMesh(Mesh*) const;
    void computeFacetNormals(Mesh*) const;
    void computeAdjacencies(Mesh*) const;
    void computeLightingNormals(Mesh*) const;
    void drawMesh(const Mesh&) const;
  private:
    Mesh mesh;
};

IApp* CreateApp() { return (IApp*) new CApp; }

CApp::CApp()
{
    const int subdivisionCount = 3;
    initGL();
    createMesh(&mesh, gizmo);
    for (int i = 0; i < subdivisionCount; ++i)
    {
        computeAdjacencies(&mesh);
        subdivideMesh(&mesh);
    }
    computeAdjacencies(&mesh);
    computeFacetNormals(&mesh);
    computeLightingNormals(&mesh);
}


void CApp::Update()
{
}


void CApp::initGL() const
{
    glClearColor(0.125, 0.25, 0.5, 1.0);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    //glEnable(GL_CULL_FACE);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    glEnable(GL_DEPTH_TEST);
    //glOrtho(-1, 1, -1, 1, -100, 100);
    glFrustum(-1, 1, -1, 1, 1.5, 20);
    glMatrixMode(GL_MODELVIEW);
    glTranslatef(0, 0, -1.6f);

    float yellow[] = {0.9f, 0.75f, 0, 1};
    float white[] = {1, 1, 1, 1};

    glMaterialfv(GL_FRONT, GL_DIFFUSE, yellow);
    glMaterialfv(GL_FRONT, GL_SPECULAR, white);
    glMaterialf(GL_FRONT, GL_SHININESS, 7.5f);
}


void CApp::Draw() const
{
    static double start = osGetSeconds();
    double theta = (osGetSeconds() - start) * 10;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glPushMatrix();
    glTranslatef(0, 0, -1);
    glRotated(theta, 0, 0, 1);
    glRotated(theta, 1, 0, 0);
    drawMesh(mesh);
    glPopMatrix();
}


void CApp::OnKey(char c)
{
}


void CApp::createMesh(Mesh* mesh, const MeshEgg& egg) const
{
    mesh->points.resize(egg.pointCount);
    mesh->faces.resize(egg.faceCount);

    for (unsigned int index = 0; index < egg.pointCount; ++index)
    {
        mesh->points[index].position.x = egg.points[index * 3 + 0];
        mesh->points[index].position.y = egg.points[index * 3 + 1];
        mesh->points[index].position.z = egg.points[index * 3 + 2];
    }

    for (unsigned int index = 0; index < egg.faceCount; ++index)
    {
        mesh->faces[index].vertices[0].index = egg.faces[index * 4 + 0];
        mesh->faces[index].vertices[1].index = egg.faces[index * 4 + 1];
        mesh->faces[index].vertices[2].index = egg.faces[index * 4 + 2];
        mesh->faces[index].vertices[3].index = egg.faces[index * 4 + 3];
    }
}


void CApp::drawMesh(const Mesh& mesh) const
{
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glBegin(GL_QUADS);
    for (FaceList::const_iterator f = mesh.faces.begin(); f != mesh.faces.end(); ++f)
    {
        for (int corner = 0; corner < 4; ++corner)
        {
            glNormal3fv(&f->vertices[corner].lightingNormal.x);
            glVertex3fv(&mesh.points[f->vertices[corner].index].position.x);
        }
    }
    glEnd();
}


struct Edge
{
    unsigned int pointIndex;
    unsigned int faceCount;
    vec3 faceCentroidsSum;
};

typedef std::pair<unsigned int, unsigned int> PointPair;

struct EdgeComparison
{
    size_t operator()(const PointPair& key) const
    {
        if (key.first < key.second)
            return key.first + key.second * 2;

        return key.first * 2 + key.second;
    }

    // (key1, key2) maps to the same location as (key2, key1)
    bool operator() (const PointPair& key1, const PointPair& key2) const
    {
        unsigned int key10, key11, key20, key21;

        if (key1.first < key1.second)
        {
            key10 = key1.first;
            key11 = key1.second;
        }
        else
        {
            key11 = key1.first;
            key10 = key1.second;
        }

        if (key2.first < key2.second)
        {
            key20 = key2.first;
            key21 = key2.second;
        }
        else
        {
            key21 = key2.first;
            key20 = key2.second;
        }

        if (key11 == key21)
            return key10 < key20;

        return key11 < key21;
    }

    enum
    {
        bucket_size = 4,
        min_buckets = 8
    };
};

typedef stdext::hash_map<PointPair, Edge, EdgeComparison> EdgeMap;


// TODO make a new copy instead of in-place division (for storing multiple LODs)
//      also update adjacency stuff so there's no need to re-call computeAdjacencies
void CApp::subdivideMesh(Mesh* mesh) const
{
    unsigned int faceCount = mesh->faces.size();
    unsigned int newFaceCount = faceCount * 4;
    unsigned int pointCount = mesh->points.size();
    unsigned int newPointCount = pointCount + mesh->edgeCount + faceCount;
    unsigned int oldFaceCount = faceCount;
    unsigned int oldPointCount = pointCount;

    mesh->points.resize(newPointCount);
    mesh->faces.resize(newFaceCount);

    EdgeMap edgeMap;

    for (unsigned int f = 0; f < oldFaceCount; ++f)
    {
        Vertex* vertices = mesh->faces[f].vertices;

        vec3& a = mesh->points[vertices[0].index].position;
        vec3& b = mesh->points[vertices[1].index].position;
        vec3& c = mesh->points[vertices[2].index].position;
        vec3& d = mesh->points[vertices[3].index].position;

        // Create a point at the center of this face.
        unsigned int centroid = pointCount++;
        mesh->points[centroid].position = (a + b + c + d) / 4;

        // Loop over edges and create midpoints if they don't already exist.
        for (int corner = 0; corner < 4; ++corner)
        {
            PointPair pp(vertices[corner].index, vertices[(corner + 1) % 4].index);
            EdgeMap::iterator iterator = edgeMap.find(pp);
            if (iterator == edgeMap.end())
            {
                edgeMap[pp].pointIndex = pointCount;
                edgeMap[pp].faceCount = 1;
                edgeMap[pp].faceCentroidsSum = mesh->points[centroid].position;
                vec3& a = mesh->points[pp.first].position;
                vec3& b = mesh->points[pp.second].position;
                mesh->points[pointCount++].position = (a + b) / 2;
            }
            else
            {
                edgeMap[pp].faceCount++;
                edgeMap[pp].faceCentroidsSum += mesh->points[centroid].position;
            }
            mesh->points[pp.first].faceCentroidsSum += mesh->points[centroid].position;
        }

        // Create 4 new faces.  The 4th face replaces the current face.
        for (int corner = 0; corner < 4; ++corner)
        {
            PointPair edge0(vertices[corner % 4].index, vertices[(corner + 1) % 4].index);
            PointPair edge1(vertices[(corner + 1) % 4].index, vertices[(corner + 2) % 4].index);
            Face& face = (corner == 3) ? mesh->faces[f] : mesh->faces[faceCount++];
            face.vertices[1].index = vertices[(corner + 1) % 4].index;
            face.vertices[0].index = edgeMap[edge0].pointIndex;
            face.vertices[2].index = edgeMap[edge1].pointIndex;
            face.vertices[3].index = centroid;
        }
    }

    // Move the old vertices
    for (unsigned int p = 0; p < oldPointCount; ++p)
    {
        AdjacencySet& verts = mesh->points[p].adjacentVerts;
        AdjacencyList& faces = mesh->points[p].adjacentFaces;
        vec3& faceCentroidsSum = mesh->points[p].faceCentroidsSum;

        vec3 edgeMidpointsSum(0, 0, 0);
        unsigned int creaseCount = 0;
        vec3 creaseSum(0, 0, 0);
        for (AdjacencySet::iterator v = verts.begin(); v != verts.end(); ++v)
        {
            PointPair edge(p, *v);
            edgeMidpointsSum += mesh->points[edgeMap[edge].pointIndex].position;
            if (edgeMap[edge].faceCount == 1)
            {
                creaseCount++;
                creaseSum += mesh->points[*v].position;
            }
        }

        if (creaseCount > 2)
            osError("corners not yet supported");

        if (creaseCount == 1)
            osError("darts not yet supported");

        if (creaseCount == 2)
        {
            vec3 R = creaseSum / (float) creaseCount;
            vec3& S = mesh->points[p].position;
            mesh->points[p].position = R * 0.25f + S * 0.75f;
            continue;
        }

        vec3 Q = faceCentroidsSum / (float) faces.size();
        vec3 R = edgeMidpointsSum / (float) verts.size();
        vec3& S = mesh->points[p].position;
        float n = (float) verts.size();

        mesh->points[p].position = (Q + 2 * R + S * (n - 3)) / n;
    }

    // Move the new edge vertices.
    for (EdgeMap::iterator i = edgeMap.begin(); i != edgeMap.end(); ++i)
    {
        Edge& edge = i->second;
        if (edge.faceCount > 1)
        {
            mesh->points[edge.pointIndex].position += edge.faceCentroidsSum;
            mesh->points[edge.pointIndex].position /= (float) edge.faceCount + 1.0f;
        }
    }

    assert(pointCount == newPointCount);
    assert(faceCount == newFaceCount);
}



void CApp::computeFacetNormals(Mesh* mesh) const
{
    // assume that all quads are planar
    for (FaceList::iterator f = mesh->faces.begin(); f != mesh->faces.end(); ++f)
    {
        vec3& a = mesh->points[f->vertices[0].index].position;
        vec3& b = mesh->points[f->vertices[1].index].position;
        vec3& c = mesh->points[f->vertices[2].index].position;
        f->facetNormal = cross(c - a, b - a).normalized();
    }
}


void CApp::computeAdjacencies(Mesh* mesh) const
{
    // Clear out the adjacency information.  Needed for re-entrancy.
    for (unsigned int p = 0; p < mesh->points.size(); ++p)
    {
        Point& point = mesh->points[p];
        point.adjacentFaces.clear();
        point.adjacentVerts.clear();
        point.faceCentroidsSum = vec3(0, 0, 0);
    }

    // Walk through each face and add it to the face lists of its vertices.
    for (unsigned int f = 0; f < mesh->faces.size(); ++f)
    {
        Vertex* vertices = mesh->faces[f].vertices;
        for (int corner = 0; corner < 4; ++corner)
        {
            mesh->points[vertices[corner].index].adjacentFaces.push_back(f);
            mesh->points[vertices[(corner + 1) % 4].index].adjacentVerts.insert(vertices[corner].index);
            mesh->points[vertices[(corner + 1) % 4].index].adjacentVerts.insert(vertices[(corner + 2) % 4].index);
        }
    }

    // Classic graph theory.  The number of edges is half the sum of all valences.
    unsigned int degreeSum = 0;
    for (unsigned int p = 0; p < mesh->points.size(); ++p)
        degreeSum += mesh->points[p].adjacentVerts.size();
    mesh->edgeCount = degreeSum / 2;
}


void CApp::computeLightingNormals(Mesh* mesh) const
{
    for (unsigned int f = 0; f < mesh->faces.size(); ++f)
    {
        Vertex* vertices = mesh->faces[f].vertices;
        for (int corner = 0; corner < 4; ++corner)
        {
            Point& point = mesh->points[vertices[corner].index];

            vec3 sum(0, 0, 0);
            for (AdjacencyList::const_iterator n = point.adjacentFaces.begin(); n != point.adjacentFaces.end(); ++n)
                sum += mesh->faces[*n].facetNormal;

            vertices[corner].lightingNormal = sum.normalized();
        }
    }
}
