// File:      Vector.cpp
// Author:    Philip Rideout
// Brief:     Basic vector math.

#include "Platform.h"
#include <cmath>
#include <algorithm>
#include "Vector.h"

float vec3::magnitude() const
{
    return sqrtf(mag2());
}

float vec3::mag2() const
{
    return x*x + y*y + z*z;
}

void vec3::unitize()
{
    float m = magnitude();
    if (!m)
        return;
    x /= m;
    y /= m;
    z /= m;
}

vec3 vec3::normalized() const
{
    float m = magnitude();
    if (!m)
        return *this;
    return vec3(x / m, y / m, z / m);
}

float vec2::magnitude() const
{
    return sqrtf(mag2());
}

float vec2::mag2() const
{
    return x*x + y*y;
}

void vec2::unitize()
{
    float m = magnitude();
    if (!m)
        return;
    x /= m;
    y /= m;
}

float dot(const vec3& a, const vec3& b)
{
    return a.x*b.x + a.y*b.y + a.z*b.z;
}

vec3 cross(const vec3& u, const vec3& v)
{
    return vec3(u.y*v.z-u.z*v.y, u.z*v.x-u.x*v.z, u.x*v.y-u.y*v.x);
}

mat4 operator*(const mat4& a, const mat4& b) // TODO stop cheating
{
    mat4 retval;
    glPushMatrix();
    glLoadMatrixf(a.data);
    glMultMatrixf(b.data);
    glGetFloatv(GL_MODELVIEW_MATRIX, retval.data);
    glPopMatrix();
    return retval;
}

vec2 operator*(const vec2& v, const mat4& m)
{
    float x = m.data[0] * v.x + m.data[4] * v.y + m.data[12];
    float y = m.data[1] * v.x + m.data[5] * v.y + m.data[13];
    return vec2(x, y);
}

vec3 blend(const vec3& v, const mat4& m0, const mat4& m1, float w0, float w1)
{
    return v * m0 * w0 + v * m1 * w1;
}

vec3 operator*(const vec3& v, const mat4& m)
{
    float x = m.data[0] * v.x + m.data[4] * v.y + m.data[8] * v.z  + m.data[12];
    float y = m.data[1] * v.x + m.data[5] * v.y + m.data[9] * v.z  + m.data[13];
    float z = m.data[2] * v.x + m.data[6] * v.y + m.data[10] * v.z + m.data[14];
    float w = m.data[3] * v.x + m.data[7] * v.y + m.data[11] * v.z + m.data[15];
    vec3 vPrime(x, y, z);
    if (w)
        return vPrime / w;
    return vPrime;
}

vec3 operator*(float scale, const vec3& v)
{
    return v * scale;
}

void mat4::identity()
{
    for (int i = 0; i < 4; ++i)
        for (int j = 0; j < 4; ++j)
            data[i*4 + j] = (i == j) ? 1.0f : 0.0f;
}

bool vec3::operator<(const vec3& v) const
{
    if (x != v.x)
        return x < v.x;
    if (y != v.y)
        return y < v.y;
    return z < v.z;
}

float& vec3::operator[](int index)
{
    if (index == Z) return z;
    if (index == Y) return y;
    return x;
}

const float& vec3::operator[](int index) const
{
    if (index == Z) return z;
    if (index == Y) return y;
    return x;
}

int& ivec3::operator[](int index)
{
    if (index == Z) return z;
    if (index == Y) return y;
    return x;
}

const int& ivec3::operator[](int index) const
{
    if (index == Z) return z;
    if (index == Y) return y;
    return x;
}

void mat4::invert() // TODO cleanup
{
    int i, j, k, swap;
    float t;
    mat4 tempSrc;
    mat4 tempInverse;

    tempInverse.identity();
    tempSrc = *this;

    for (i = 0; i < 4; i++)
    {
        //
        // Look for largest element in column
        //
        swap = i;
        for (j = i + 1; j < 4; j++)
        {
            if (fabs(tempSrc.rows[j][i]) > fabs(tempSrc.rows[swap][i]))
            {
                swap = j;
            }
        }

        //
        // Swap rows if necessary
        //
        if (swap != i)
        {
            for (k = 0; k < 4; k++)
            {
                t = tempSrc.rows[i][k];
                tempSrc.rows[i][k] = tempSrc.rows[swap][k];
                tempSrc.rows[swap][k] = t;

                t = tempInverse.rows[i][k];
                tempInverse.rows[i][k] = tempInverse.rows[swap][k];
                tempInverse.rows[swap][k] = t;
            }
        }

        //
        // Check for singular matrix.
        // If there is no non-zero pivot, the matrix is singular, which typically shouldn't happen.
        //
        if (!tempSrc.rows[i][i])
        {
            *this = tempInverse;
            return;
        }

        t = 1.0F / tempSrc.rows[i][i];
        for (k = 0; k < 4; k++)
        {
            tempSrc.rows[i][k] *= t;
            tempInverse.rows[i][k] *= t;
        }

        for (j = 0; j < 4; j++)
        {
            if (j != i)
            {
                t = tempSrc.rows[j][i];

                for (k = 0; k < 4; k++)
                {
                    tempSrc.rows[j][k] -= tempSrc.rows[i][k] * t;
                    tempInverse.rows[j][k] -= tempInverse.rows[i][k] * t;
                }
            }
        }
    }

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            rows[i][j] = tempInverse.rows[i][j];
}


mat4 mat4::inverse() const // TODO cleanup
{
    int i, j, k, swap;
    float t;
    mat4 tempSrc;
    mat4 tempInverse;

    tempInverse.identity();
    tempSrc = *this;

    for (i = 0; i < 4; i++)
    {
        //
        // Look for largest element in column
        //
        swap = i;
        for (j = i + 1; j < 4; j++)
        {
            if (fabs(tempSrc.rows[j][i]) > fabs(tempSrc.rows[swap][i]))
            {
                swap = j;
            }
        }

        //
        // Swap rows if necessary
        //
        if (swap != i)
        {
            for (k = 0; k < 4; k++)
            {
                t = tempSrc.rows[i][k];
                tempSrc.rows[i][k] = tempSrc.rows[swap][k];
                tempSrc.rows[swap][k] = t;

                t = tempInverse.rows[i][k];
                tempInverse.rows[i][k] = tempInverse.rows[swap][k];
                tempInverse.rows[swap][k] = t;
            }
        }

        //
        // Check for singular matrix.
        // If there is no non-zero pivot, the matrix is singular, which typically shouldn't happen.
        //
        if (!tempSrc.rows[i][i])
            return tempInverse;

        t = 1.0F / tempSrc.rows[i][i];
        for (k = 0; k < 4; k++)
        {
            tempSrc.rows[i][k] *= t;
            tempInverse.rows[i][k] *= t;
        }

        for (j = 0; j < 4; j++)
        {
            if (j != i)
            {
                t = tempSrc.rows[j][i];

                for (k = 0; k < 4; k++)
                {
                    tempSrc.rows[j][k] -= tempSrc.rows[i][k] * t;
                    tempInverse.rows[j][k] -= tempInverse.rows[i][k] * t;
                }
            }
        }
    }

    mat4 dest;
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            dest.rows[i][j] = tempInverse.rows[i][j];

    return dest;
}

// Given two unit vectors a and b, find the matrix that rotates from a to b.
// Based on quaternion math in "Real-Time Rendering" by Haines and Moller.
mat4 generate_rotation_matrix(const vec3& a, const vec3& b)
{
    mat4 rotation;

    if (a == b)
    {
        rotation.identity();
        return rotation;
    }

    float e = dot(a, b);
    vec3 v = cross(a, b);

    // in the degenerate case, just take a guess...
    if (!v.x && !v.y && !v.z)
    {
        rotation.identity();
        if (a.z || a.x)
        {
            rotation.data[0] = -1;
            rotation.data[10] = -1;
        }
        else if (a.y)
        {
            rotation.data[5] = -1;
            rotation.data[10] = -1;
        }
        return rotation;
    }

    float h = (1 - e) / dot(v, v);
    rotation.data[0] = e + h * v.x * v.x;
    rotation.data[1] = h * v.x * v.y - v.z;
    rotation.data[2] = h * v.x * v.z + v.y;
    rotation.data[3] = 0;

    rotation.data[4] = h * v.x * v.y + v.z;
    rotation.data[5] = e + h * v.y * v.y;
    rotation.data[6] = h * v.y * v.z - v.x;
    rotation.data[7] = 0;

    rotation.data[8] = h * v.x * v.z - v.y;
    rotation.data[9] = h * v.y * v.z + v.x;
    rotation.data[10] = e + h * v.z * v.z;
    rotation.data[11] = 0;

    rotation.data[12] = 0;
    rotation.data[13] = 0;
    rotation.data[14] = 0;
    rotation.data[15] = 1;
    return rotation;
}

Axis vec3::max_coord() const
{
    return ((x > y) ? ((x > z) ? X : Z) : ((y > z) ? Y : Z));
}

vec3 vec3::abs() const
{
    return vec3(fabsf(x), fabsf(y), fabsf(z));
}

Axis ivec2::max_coord() const
{
    return (x > y) ? X : Y;
}

ivec2 ivec2::abs() const
{
    return ivec2(::abs(x), ::abs(y));
}

vec3 reflect(const vec3& a, const vec3& b)
{
    float k = dot(a, b) * -2.0f;
    return k * b + a;
}

vec2 latlong(vec3 v)
{
    v.unitize();
    double theta = acosf(v.y);
    double phi = atan2(-v.z, v.x) + 3 * M_PI / 2;
    double s = phi / (2 * M_PI);
    double t = theta / M_PI;
    return vec2((float) s, (float) t);
}

void glVertex(const vec3& v)
{
    glVertex3fv((float*) &v);
}

void glVertex(const vec2& v)
{
    glVertex2fv((float*) &v);
}

void glVertex(const ivec2& v)
{
    glVertex2iv((int*) &v);
}

void glNormal(const vec3& v)
{
    glNormal3fv((float*) &v);
}

void glColor(const vec3& v, float alpha)
{
    glColor4f(v.x, v.y, v.z, alpha);
}

void glTexCoord(const vec2& v)
{
    glTexCoord2f(v.x, v.y);
}
