// File:      Vector.h
// Author:    Philip Rideout
// Brief:     Vectors and matrices.

#pragma once

enum Axis {X, Y, Z};

static const float pi = 3.14159265358979323846f;
static const float twopi = 6.28318530717958647692f;
static const float sqrt2 = 1.4142135623730950488016887242097f;

struct vec2;
struct vec3;

struct ivec2
{
    ivec2() : x(0), y(0) {}
    ivec2(int x, int y) : x(x), y(y) {}
    ivec2(const vec3&);
    ivec2(const vec2&);
    ivec2 operator-(const ivec2& v) const { return ivec2(x - v.x, y - v.y); }
    ivec2 operator+(const ivec2& v) const { return ivec2(x + v.x, y + v.y); }
    void operator+=(const ivec2& v) { x += v.x; y += v.y; }
    ivec2 operator/(int s) const { return ivec2(x / s, y / s); }
    Axis max_coord() const;
    ivec2 abs() const;
    int x, y;
};

struct ivec3
{
    ivec3() : x(0), y(0), z(0) {}
    ivec3 operator+(const ivec3& v) const { return ivec3(v.x + x, v.y + y, v.z + z); }
    void operator+=(const ivec3& v) { x += v.x; y += v.y; z += v.z; }
    ivec3(int x, int y, int z) : x(x), y(y), z(z) {}
    int& operator[](int index);
    const int& operator[](int index) const;
    union
    {
        struct { int x, y, z; };
        struct { int i, j, k; };
    };
};

struct ivec4
{
    ivec4() : x(0), y(0), z(0), w(0) {}
    ivec4 operator+(const ivec4& v) const { return ivec4(v.x + x, v.y + y, v.z + z, v.w + w); }
    void operator+=(const ivec4& v) { x += v.x; y += v.y; z += v.z; w += v.w; }
    ivec4(int x, int y, int z, int w) : x(x), y(y), z(z), w(w) {}
    int& operator[](int index);
    const int& operator[](int index) const;
    union
    {
        struct { int x, y, z, w; };
        struct { int i, j, k, l; };
    };
};

struct vec4
{
    vec4() {}
    vec4(float x, float y, float z, float w) : x(x), y(y), z(z), w(w) {}
    vec4(const vec3& v);
    vec4(const vec3& v, float f);
    void operator+=(const vec4& v) { x += v.x; y += v.y; z += v.z; w += v.w; }

    void blend_color(const vec4& source)
    {
        red = source.red * source.alpha + red * (1 - source.alpha);
        green = source.green * source.alpha + green * (1 - source.alpha);
        blue = source.blue * source.alpha + blue * (1 - source.alpha);
        alpha = source.alpha * source.alpha + alpha * (1 - source.alpha);
        if (red > 1) red = 1;
        if (green > 1) green = 1;
        if (blue > 1) blue = 1;
        if (alpha > 1) alpha = 1;
    }

    union
    {
        struct { float x, y, z, w; };
        struct { float r, g, b, a; };
        struct { float red, green, blue, alpha; };
    };
};

struct vec3
{
    vec3() : x(0), y(0), z(0) {}
    vec3(const vec4& v) : x(v.x), y(v.y), z(v.z) {}
    vec3(float x, float y, float z) : x(x), y(y), z(z) {}
    vec3 operator-(const vec3& v) const { return vec3(x - v.x, y - v.y, z - v.z); }
    vec3 operator+(const vec3& v) const { return vec3(x + v.x, y + v.y, z + v.z); }
    vec3 operator-() const { return vec3(-x, -y, -z); }
    vec3 operator/(float f) const { return vec3(x / f, y / f, z / f); }
    vec3 operator*(float f) const { return vec3(x * f, y * f, z * f); }
    vec3 operator*(const vec3& v) const { return vec3(x * v.x, y * v.y, z * v.z); }
    void operator*=(float scale) { x *= scale; y *= scale; z *= scale; }
    void operator/=(float f) { x /= f; y /= f; z /= f; }
    void operator+=(const vec3& v) { x += v.x; y += v.y; z += v.z; }
    void operator-=(const vec3& v) { x -= v.x; y -= v.y; z -= v.z; }
    bool operator==(const vec3& v) const { return x == v.x && y == v.y && z == v.z; }
    float& operator[](int index);
    const float& operator[](int index) const;
    void set(float* f) const { f[0] = x; f[1] = y; f[2] = z; }
    float magnitude() const;
    float mag2() const;
    Axis max_coord() const;
    vec3 abs() const;
    void unitize(); // TODO "normalize(v)" would be more consistent with GLSL, also length(v)
    vec3 normalized() const;
    union
    {
        struct { float x, y, z; };
        struct { float r, g, b; };
    };
    bool operator<(const vec3& v) const;
};

struct vec2
{
    vec2() : x(0), y(0) {}
    vec2(float x, float y) : x(x), y(y) {}
    vec2(int x, int y) : x((float) x), y((float) y) {}
    vec2(float x) : x(x), y(x) {}
    vec2(const float p[]) : x(p[0]), y(p[1]) {}
    vec2(const vec3& v) : x(v.x), y(v.y) {}
    vec2 operator/(float f) const { return vec2(x / f, y / f); }
    vec2 operator*(float scale) const { return vec2(x * scale, y * scale); }
    vec2 operator*(const vec2& v) const { return vec2(x * v.x, y * v.y); }
    vec2 operator+(vec2 v) const { return vec2(v.x + x, v.y + y); }
    vec2 operator-(vec2 v) const { return vec2(x - v.x, y - v.y); }
    void operator+=(vec2 v) { x += v.x; y += v.y; }
    void operator*=(float scale) { x *= scale; y *= scale; }
    void operator/=(float scale) { x /= scale; y /= scale; }
    union { float x, u, s; };
    union { float y, v, t; };
    void flip() { float temp = x; x = y; y = temp; }
    float magnitude() const;
    float mag2() const;
    void unitize();
};

struct mat2 { float data[4]; };

struct mat3 { float data[9]; };

struct mat4
{
    union
    {
        float data[16];
        float rows[4][4];
    };
    void identity();
    void invert();
    mat4 inverse() const;
};

mat4 operator*(const mat4& a, const mat4& b);

vec3 blend(const vec3& v, const mat4& m0, const mat4& m1, float w0, float w1);
vec3 cross(const vec3& a, const vec3& b);
float dot(const vec3& a, const vec3& b);
vec2 operator*(const vec2& v, const mat4& m);
vec3 operator*(const vec3& v, const mat4& m);
vec3 operator*(float scale, const vec3& v);
mat4 generate_rotation_matrix(const vec3&, const vec3&);

inline ivec2::ivec2(const vec3& v) { x = (int) v.x; y = (int) v.y; }
inline ivec2::ivec2(const vec2& v) { x = (int) v.x; y = (int) v.y; }
inline vec4::vec4(const vec3& v) : x(v.x), y(v.y), z(v.z), w(1.0f) {}
inline vec4::vec4(const vec3& v, float f) : x(v.x), y(v.y), z(v.z), w(f) {}

vec3 reflect(const vec3& a, const vec3& b);
vec2 latlong(vec3 v);

void glVertex(const ivec2& v);
void glVertex(const vec2& v);
void glVertex(const vec3& v);
void glNormal(const vec3& v);
void glColor(const vec3& v, float alpha = 1);
void glTexCoord(const vec2& v);
