// File:      WinMain.cpp
// Author:    Philip Rideout
// Brief:     Application entry point for Win32 platforms.

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <cstdio>
#include <cstdlib>
#include <GL/glew.h>
#include <GL/wglew.h>
#include "App.h"
#define MAX_ERROR_LENGTH 128

int osGetScreenWidth();
int osGetScreenHeight();
void osSetWindowTitle(const char*);
void osError(const char* message);
void CreateOpenGLWindow(LPCTSTR name, HINSTANCE hInstance, HDC* hDC, HGLRC* hRC, HWND* hWnd, int pixelFormat = -1);
void CreateFSAAWindow(LPCTSTR name, HINSTANCE hInstance, HDC* hDC, HGLRC* hRC, HWND* hWnd);
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow);
LRESULT CALLBACK WinProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void Quit(HWND hWnd);
void glCheckError(const char* call);


int osGetScreenWidth() { return GetSystemMetrics(SM_CXSCREEN); }
int osGetScreenHeight() { return GetSystemMetrics(SM_CYSCREEN); }
void osSetWindowTitle(const char* txt) { SetWindowText(GetActiveWindow(), txt); }
double osGetSeconds() { return (double) GetTickCount() / 1000; }


void osError(const char* message)
{
    MessageBox(0, message, "Error", MB_ICONERROR);
#ifdef _DEBUG
    __debugbreak();
#else
    exit(EXIT_FAILURE);
#endif
}


void CreateOpenGLWindow(LPCTSTR name, HINSTANCE hInstance, HDC* hDC, HGLRC* hRC, HWND* hWnd, int pixelFormat)
{
    // Create the window.
    RECT rect;

    if (const bool fullscreen = false)
        SetRect(&rect, 0, 0, osGetScreenWidth(), osGetScreenHeight());
    else
        SetRect(&rect, 0, 0, 512, 512);

    DWORD dwStyle = WS_CAPTION | WS_SYSMENU | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    DWORD dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
    AdjustWindowRectEx(&rect, dwStyle, FALSE, dwExStyle);
    int width = rect.right - rect.left;
    int height = rect.bottom - rect.top;
    *hWnd = CreateWindowEx(0, name, name, dwStyle, 0, 0, width, height, 0, 0, hInstance, 0);

    // Create the GL context.
    PIXELFORMATDESCRIPTOR pfd;
    ZeroMemory(&pfd, sizeof(pfd));
    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 32;
    pfd.cDepthBits = 8;
    pfd.cStencilBits = 1;
    pfd.iLayerType = PFD_MAIN_PLANE;

    *hDC = GetDC(*hWnd);
    if (pixelFormat == -1)
        pixelFormat = ChoosePixelFormat(*hDC, &pfd);

    SetPixelFormat(*hDC, pixelFormat, &pfd);
    *hRC = wglCreateContext(*hDC);
    wglMakeCurrent(*hDC, *hRC);
}


void CreateFSAAWindow(LPCTSTR name, HINSTANCE hInstance, HDC* hDC, HGLRC* hRC, HWND* hWnd)
{
    PROC proc = wglGetProcAddress("wglChoosePixelFormatARB");
    PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC) proc;
    if (!wglChoosePixelFormatARB)
        osError("Could not load function pointer for 'wglChoosePixelFormatARB'.  Is your driver properly installed?");

    unsigned int numFormats;
    int iAttributes[] =
    {
        WGL_SAMPLES_ARB, 16,
        WGL_SAMPLE_BUFFERS_ARB, GL_TRUE,
        WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
        WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
        WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
        WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
        WGL_RED_BITS_ARB, 8,
        WGL_GREEN_BITS_ARB, 8,
        WGL_BLUE_BITS_ARB, 8,
        WGL_ALPHA_BITS_ARB, 8,
        WGL_DEPTH_BITS_ARB, 24,
        WGL_STENCIL_BITS_ARB, 1,
        WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
        0, 0
    };

    int& sampleCount = iAttributes[1];
    int& useSampleBuffer = iAttributes[3];
    int pixelFormat = -1;

    // Try fewer and fewer samples per pixel till we find one that is supported.
    while (pixelFormat <= 0 && sampleCount >= 0)
    {
        wglChoosePixelFormatARB(*hDC, iAttributes, 0, 1, &pixelFormat, &numFormats);
        sampleCount--;
        if (sampleCount <= 1)
            useSampleBuffer = GL_FALSE;
    }

    // Win32 allows the pixel format to be set only once per window, so destroy and re-create the window.
    if (pixelFormat != -1)
    {
        DestroyWindow(*hWnd);
        CreateOpenGLWindow(name, hInstance, hDC, hRC, hWnd, pixelFormat);
    }
}


int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    // Register the window class with some extra bytes for storing an App pointer.
    LPCTSTR name = "CatmullClark";
    WNDCLASS wndClass =
    {
        CS_OWNDC, WinProc, 0, sizeof(IApp*), 
        hInstance, 0, LoadCursor(0, IDC_ARROW), 0, 0, name
    };
    RegisterClass(&wndClass);

    // Create the window and create the GL context.
    HDC hDC;
    HGLRC hRC;
    HWND hWnd;
    CreateOpenGLWindow(name, hInstance, &hDC, &hRC, &hWnd);
    CreateFSAAWindow(name, hInstance, &hDC, &hRC, &hWnd);

    // Initialize the extension wrangler and check the OpenGL version.
    glewInit();

    const GLubyte* version = glGetString(GL_VERSION);
    if (!version)
        osError("OpenGL is not available.");

    int major, minor;
    _snscanf((const char*) version, 8, "%d.%d", &major, &minor);
    if (major < 2)
        osError("OpenGL 2.0 is required.");

    // Create the application object and add it to the window's private user data.
    IApp* app = CreateApp();
    SetWindowLongPtr(hWnd, GWLP_USERDATA, PtrToLong(app));

    // Main message loop.
    MSG msg;
    msg.message = WM_NULL;
    while (msg.message != WM_QUIT)
    {
        if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
        else
        {
            app->Update();
            app->Draw();
            SwapBuffers(hDC);
        }
    }

    return EXIT_SUCCESS;
}


LRESULT CALLBACK WinProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    IApp* app = (IApp*) LongToPtr(GetWindowLongPtr(hWnd, GWLP_USERDATA));

    switch (uMsg)
    {
        case WM_CHAR:
            if (wParam == VK_ESCAPE || wParam == 'q' || wParam == 'Q' || wParam == 'x' || wParam == 'X')
            {
                Quit(hWnd);
                app = 0;
            }
            else
            {
                app->OnKey((char) wParam);
            }
            break;

        case WM_CLOSE:
            Quit(hWnd);
            app = 0;
            break;
    }

    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


void Quit(HWND hWnd)
{
    IApp* app = (IApp*) LongToPtr(GetWindowLongPtr(hWnd, GWLP_USERDATA));
    delete app;
    SetWindowLongPtr(hWnd, GWLP_USERDATA, 0);

    HGLRC hRC = wglGetCurrentContext();
    wglDeleteContext(hRC);

    HDC hDC = GetDC(hWnd);
    DeleteDC(hDC);

    DestroyWindow(hWnd);
    PostQuitMessage(0);
}


void glCheckError(const char* call)
{
    char message[MAX_ERROR_LENGTH];
    char enums[][20] =
    {
        "invalid enumeration", // GL_INVALID_ENUM
        "invalid value",       // GL_INVALID_VALUE
        "invalid operation",   // GL_INVALID_OPERATION
        "stack overflow",      // GL_STACK_OVERFLOW
        "stack underflow",     // GL_STACK_UNDERFLOW
        "out of memory"        // GL_OUT_OF_MEMORY
    };

    GLenum errcode = glGetError();
    if (errcode == GL_NO_ERROR)
        return;

    errcode -= GL_INVALID_ENUM;
    _snprintf(message, MAX_ERROR_LENGTH, "OpenGL %s in '%s'", enums[errcode], call);
    osError(message);
}
